import abc
import asyncio
from typing import *

from llbase import llsd


class AbstractLEAPProtocol(abc.ABC):
    """Interface for a class representing communication with a LEAP peer"""

    closed: bool

    @abc.abstractmethod
    def close(self) -> None:
        pass

    @abc.abstractmethod
    def write_message(self, pump: str, data: Any) -> None:
        pass

    @abc.abstractmethod
    async def read_message(self) -> Dict:
        pass


class LEAPProtocol(AbstractLEAPProtocol):
    """Wrapper for communication with a LEAP peer over an asyncio reader/writer pair"""

    def __init__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        self._reader = reader
        self._writer = writer
        self._parser = llsd.LLSDNotationParser()
        self._formatter = llsd.LLSDNotationFormatter()
        self._drain_task = None

    @property
    def closed(self) -> bool:
        return self._writer.is_closing() or self._reader.at_eof()

    def close(self):
        if not self._writer.is_closing():
            self._writer.write_eof()
            self._writer.close()

    def write_message(self, pump: str, data: Any) -> None:
        assert not self._writer.is_closing()
        ser = self._formatter.format({"pump": pump, "data": data})
        payload = bytearray(str(len(ser)).encode("utf8"))
        payload.extend(b":")
        payload.extend(ser)
        self._writer.write(payload)
        # We're in sync context, we need to schedule draining the socket, which is async.
        # If a drain is already scheduled then we don't need to reschedule.
        if not self._drain_task:
            self._drain_task = asyncio.create_task(self._drain_soon())

    async def _drain_soon(self) -> None:
        self._drain_task = None
        await self._writer.drain()

    async def read_message(self) -> Dict:
        assert not self._reader.at_eof()

        # Length is everything up until the first colon we see, stripping the colon off.
        length = int((await self._reader.readuntil(b":"))[:-1].decode("utf8"))
        if length > 0xFFFFFF:
            raise ValueError(f"Unreasonable LEAP payload length of {length}")
        # Everything after the colon is LLSD
        parsed = self._parser.parse(await self._reader.readexactly(length))
        if not isinstance(parsed, dict):
            raise ValueError(f"Expected LEAP message to be a dict, got {parsed!r}")
        return parsed


__all__ = [
    "AbstractLEAPProtocol",
    "LEAPProtocol",
]



"""
Test that LEAP scripts handle STDIO operations correctly
"""

import asyncio
import sys
import unittest

import outleap

TEST_CODE = """
import asyncio

from outleap import LEAPClient, LLViewerControlAPI


class MockReqIDGenerator:
    def __init__(self):
        self.ctr = 0

    def __call__(self, *args, **kwargs):
        self.ctr += 1
        return self.ctr


async def amain():
    # Create a client speaking LEAP over stdin/stdout and connect it
    async with await LEAPClient.create_stdio_client() as client:
        # Use monotonically incrementing integer for reqids
        client._gen_reqid = MockReqIDGenerator()
        # Use our typed wrapper around the LLViewerControl LEAP API
        viewer_control_api = LLViewerControlAPI(client)
        # Ask for a config value and print it in the viewer logs
        await viewer_control_api.get("Global", "StatsPilotFile")

loop = asyncio.get_event_loop_policy().get_event_loop()
loop.run_until_complete(amain())
"""


class STDIOTests(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.proc = await asyncio.create_subprocess_exec(
            sys.executable,
            "-c",
            TEST_CODE,
            stdout=asyncio.subprocess.PIPE,
            stdin=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        self.protocol = outleap.LEAPProtocol(self.proc.stdout, self.proc.stdin)

    def _write_hello(self):
        self.protocol.write_message(
            "reply_pump",
            {
                "command": "cmd_pump",
            },
        )

    async def _assert_normal_exit(self):
        await asyncio.wait_for(self.proc.wait(), 2.0)
        # We want to know if any warnings about dangling tasks or such were printed
        self.assertEqual(b"", await self.proc.stderr.read())
        self.assertEqual(0, self.proc.returncode)

    async def test_connect_and_reply(self):
        self._write_hello()
        msg = await self.protocol.read_message()
        self.assertEqual("LLViewerControl", msg["pump"])
        self.protocol.write_message(
            "reply_pump",
            {
                "value": "whatever",
                "reqid": 1,
            },
        )
        await self._assert_normal_exit()

    async def asyncTearDown(self) -> None:
        if self.proc.returncode is None:
            self.proc.kill()



import asyncio

from . import BaseClientTest


class ProtocolTests(BaseClientTest):
    async def test_connect(self):
        self._write_welcome()
        await self.client.connect()
        self.assertTrue(self.client.connected)

    async def test_disconnect(self):
        self._write_welcome()
        await self.client.connect()
        self.client.disconnect()
        self.assertFalse(self.client.connected)

    async def test_scoped_connect(self):
        self._write_welcome()
        self.assertFalse(self.client.connected)
        async with self.client:
            self.assertTrue(self.client.connected)

        self.assertFalse(self.client.connected)

    async def test_connect_bad_welcome(self):
        # Empty welcome message
        self.protocol.inbound_messages.put_nowait({})
        with self.assertRaises(KeyError):
            await self.client.connect()

    async def test_post(self):
        self._write_welcome()
        await self.client.connect()
        self.client.post("foopump", {"bar": 1}, expect_reply=False)  # noqa
        self.assertDictEqual({"pump": "foopump", "data": {"bar": 1}}, self.protocol.sent_messages[-1])

    async def test_post_bad_data(self):
        self._write_welcome()
        await self.client.connect()
        # Must post a dict if you expect a reply
        with self.assertRaises(ValueError):
            self.client.post("foopump", "foo", expect_reply=True)  # noqa

    async def test_void_command(self):
        self._write_welcome()
        await self.client.connect()
        self.client.void_command("foopump", "baz", {"bar": 1})
        self.assertDictEqual(
            {"pump": "foopump", "data": {"op": "baz", "bar": 1}}, self.protocol.sent_messages[-1]
        )

    async def test_command(self):
        self._write_welcome()
        await self.client.connect()
        fut = self.client.command("foopump", "baz", {"bar": 1})
        # Verify that the side effect of sending the message happens _before_ we await.
        last_message = self.protocol.sent_messages[-1]
        expected_msg = {
            "pump": "foopump",
            "data": {"op": "baz", "bar": 1, "reply": "reply_pump", "reqid": 1},
        }
        self.assertDictEqual(expected_msg, last_message)

        # The reply hasn't come in yet, future should still be pending
        done, pending = await asyncio.wait([fut], timeout=0.01)
        self.assertEqual(set(), done)

        # Pretend a reply came in
        self._write_reply(1, {"foo": 1})
        self.assertEqual({"foo": 1}, await asyncio.wait_for(fut, timeout=0.01))

    async def test_disconnect_pending_command(self):
        self._write_welcome()
        async with self.client:
            fut = self.client.command("foo", "bar", {})

        with self.assertRaises(asyncio.CancelledError):
            await fut

    async def test_handle_bad_message(self):
        self._write_welcome()
        async with self.client:
            self.assertFalse(self.client.handle_message("foo"))

    async def test_handle_reply_not_dict(self):
        self._write_welcome()
        async with self.client:
            self.assertFalse(
                self.client.handle_message(
                    {
                        "pump": "reply_pump",
                        "data": "foo",
                    }
                )
            )

    async def test_handle_reply_no_reqid(self):
        self._write_welcome()
        async with self.client:
            self.assertFalse(
                self.client.handle_message(
                    {
                        "pump": "reply_pump",
                        "data": {},
                    }
                )
            )

    async def test_listen(self):
        self._write_welcome()
        await self.client.connect()
        listen_fut = self.client.listen("SomeState")
        # Pretend a reply came in allowing the listen
        self._write_reply(1)
        listener = await listen_fut

        self.protocol.inbound_messages.put_nowait(
            {
                "pump": "SomeState",
                "data": "hi",
            }
        )

        msg = await listener.get()
        self.assertEqual("hi", msg)

        # Done, unregister the listen
        stop_listen_fut = self.client.stop_listening(listener)
        # Pretend a reply came in stopping the listen
        self._write_reply(2)
        await stop_listen_fut

    async def test_listen_shutdown(self):
        self._write_welcome()
        await self.client.connect()
        listen_fut = self.client.listen("SomeState")
        # Pretend a reply came in allowing the listen
        self._write_reply(1)
        listener = await listen_fut

        self.client.disconnect()
        with self.assertRaises(asyncio.CancelledError):
            await listener.get()

    async def test_listen_message_from_before_disconnect(self):
        self._write_welcome()
        await self.client.connect()
        listen_fut = self.client.listen("SomeState")
        # Pretend a reply came in allowing the listen
        self._write_reply(1)
        listener = await listen_fut

        self.protocol.inbound_messages.put_nowait(
            {
                "pump": "SomeState",
                "data": "hi",
            }
        )
        await asyncio.sleep(0)
        self.client.disconnect()

        msg = await listener.get()
        self.assertEqual("hi", msg)

    async def test_listen_ctx_mgr(self):
        self._write_welcome()
        await self.client.connect()
        # Listener registration happening from within the contextmanager is annoying to mock,
        # so we just register a manual listener first.
        listen_fut = self.client.listen("SomeState")
        # Pretend a reply came in allowing the listen
        self._write_reply(1)
        first_listener = await listen_fut

        async with self.client.listen_scoped("SomeState") as listener:
            self.protocol.inbound_messages.put_nowait(
                {
                    "pump": "SomeState",
                    "data": "hi",
                }
            )
            msg = await listener.get()
            self.assertEqual("hi", msg)

        # Done, unregister the listen
        stop_listen_fut = self.client.stop_listening(first_listener)
        # Pretend a reply came in stopping the listen
        self._write_reply(2)
        await stop_listen_fut



#!/usr/bin/env python3
import asyncio
import sys

from outleap import LEAPClient, LLViewerControlAPI


async def amain():
    # Create a client speaking LEAP over stdin/stdout and connect it
    async with await LEAPClient.create_stdio_client() as client:
        # Use our typed wrapper around the LLViewerControl LEAP API
        viewer_control_api = LLViewerControlAPI(client)
        # Ask for a config value and print it in the viewer logs
        print(await viewer_control_api.get("Global", "StatsPilotFile"), file=sys.stderr)


def main():
    loop = asyncio.get_event_loop_policy().get_event_loop()
    loop.run_until_complete(amain())


if __name__ == "__main__":
    main()



import asyncio
import os
import stat
import sys
import threading
import time
from typing import *


class HackySTDIOTransport(asyncio.Transport):
    def __init__(self):
        super().__init__()
        self.closed = False
        self.wrote_eof = False

    def can_write_eof(self) -> bool:
        return not self.closed and not self.wrote_eof

    def write_eof(self) -> None:
        assert self.can_write_eof()
        self.wrote_eof = True

    def is_closing(self) -> bool:
        return self.closed

    def is_reading(self) -> bool:
        return False

    def close(self) -> None:
        self.closed = True

    def write(self, data: Any) -> None:
        sys.stdout.buffer.write(data)
        sys.stdout.buffer.flush()


class HackySTDIOProtocol(asyncio.Protocol):
    def __init__(self):
        super().__init__()

    def data_received(self, data: bytes) -> None:
        pass

    def eof_received(self) -> Union[bool, None]:
        return False

    def resume_writing(self) -> None:
        pass

    def pause_writing(self) -> None:
        pass

    async def _drain_helper(self):
        # no-op, we flush immediately.
        pass


def _stdin_feeder(
    transport: asyncio.Transport, reader: asyncio.StreamReader, loop: asyncio.AbstractEventLoop
):
    while not transport.is_closing():
        # Read up to 1024 bytes from stdin
        data = os.read(0, 1024)
        if not data:
            time.sleep(0.0001)
            continue
        loop.call_soon_threadsafe(reader.feed_data, data)
    reader.feed_eof()


async def _make_hacky_threaded_stdio_rw() -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
    # NB: Windows has no non-blocking IO for stdin and stdout, so we have to resort to weird
    # hacks. Do blocking IO in a separate thread and feed it to a StreamReader so that we can
    # keep API symmetry across stdio & TCP LEAP code. Otherwise, we end up with an
    # un-cancellable Future like in the run_in_executor() case. We don't care if the thread
    # is still blocking trying to read when we exit, only one thing is trying to read from
    # stdin!
    # See https://stackoverflow.com/questions/31510190/aysncio-cannot-read-stdin-on-windows
    #
    # TODO: Currently we also use this if we're reading from a non-pipe file descriptor on POSIX
    #  platforms, but that's probably unnecessary.
    loop = asyncio.get_event_loop()
    reader = asyncio.StreamReader()
    protocol = HackySTDIOProtocol()
    transport = HackySTDIOTransport()
    writer = asyncio.StreamWriter(transport, protocol, reader, loop)

    # Runs forever and dies with the process.
    t = threading.Thread(target=_stdin_feeder, args=(transport, reader, loop), daemon=True)
    t.daemon = True
    t.start()
    return reader, writer


def _is_unix_pipe(fileno: int) -> bool:
    mode = os.fstat(fileno).st_mode
    return stat.S_ISFIFO(mode) or stat.S_ISSOCK(mode) or stat.S_ISCHR(mode)


async def connect_stdin_stdout() -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
    """
    Get a StreamReader and StreamWriter for stdin and stdout, respectively
    """
    # Windows (and probably cygwin) need special logic for non-blocking use of STDIO.
    need_stdin_hack = any(sys.platform.startswith(x) for x in ("cygwin", "win32"))
    if not need_stdin_hack:
        # We also need it if stdin isn't actually a pipe, the pipe will fail to connect!
        if not _is_unix_pipe(sys.stdin.fileno()) or not _is_unix_pipe(sys.stdout.fileno()):
            need_stdin_hack = True
    if need_stdin_hack:
        return await _make_hacky_threaded_stdio_rw()
    loop = asyncio.get_event_loop()
    reader = asyncio.StreamReader()
    protocol = asyncio.StreamReaderProtocol(reader)
    await loop.connect_read_pipe(lambda: protocol, sys.stdin)
    w_transport, w_protocol = await loop.connect_write_pipe(asyncio.streams.FlowControlMixin, sys.stdout)
    writer = asyncio.StreamWriter(w_transport, w_protocol, reader, loop)
    return reader, writer


__all__ = [
    "connect_stdin_stdout",
]



import asyncio
import unittest
from typing import *

import outleap


class MockTransport(asyncio.Transport):
    def __init__(self):
        super().__init__()
        self.written_data = bytearray()
        self.closed = False
        self.wrote_eof = False

    def can_write_eof(self) -> bool:
        return not self.closed and not self.wrote_eof

    def write_eof(self) -> None:
        assert self.can_write_eof()
        self.wrote_eof = True

    def is_closing(self) -> bool:
        return self.closed

    def is_reading(self) -> bool:
        return False

    def close(self) -> None:
        self.closed = True

    def write(self, data: Any) -> None:
        self.written_data.extend(data)


class MockProtocol(asyncio.Protocol):
    def __init__(self):
        super().__init__()
        self.received_data = bytearray()

    def data_received(self, data: bytes) -> None:
        self.received_data.extend(data)

    def eof_received(self) -> Union[bool, None]:
        return False

    def resume_writing(self) -> None:
        pass

    def pause_writing(self) -> None:
        pass

    async def _drain_helper(self):
        pass


class ProtocolTests(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.loop = asyncio.get_running_loop()
        self.transport = MockTransport()
        self.protocol = MockProtocol()
        self.reader = asyncio.StreamReader()
        self.writer = asyncio.StreamWriter(self.transport, self.protocol, self.reader, self.loop)
        self.leap_protocol = outleap.LEAPProtocol(self.reader, self.writer)

    async def test_read(self):
        self.reader.feed_data(b"2:{}")
        self.assertEqual({}, await self.leap_protocol.read_message())

    async def test_write(self):
        self.leap_protocol.write_message("foo", {})
        self.assertEqual(b"24:{'pump':'foo','data':{}}", self.transport.written_data)

    async def test_read_non_dict(self):
        self.reader.feed_data(b"2:i1")
        with self.assertRaises(ValueError):
            await self.leap_protocol.read_message()

    async def test_read_too_long(self):
        payload = b"'" + (b"0" * 0xFFFFFF) + b"'"
        payload = str(len(payload)).encode("utf8") + b":" + payload
        self.reader.feed_data(payload)
        with self.assertRaisesRegex(ValueError, "length"):
            await self.leap_protocol.read_message()

    async def test_close(self):
        self.assertFalse(self.leap_protocol.closed)
        self.leap_protocol.close()
        self.assertTrue(self.leap_protocol.closed)
